/* 

  RMC - Ron's MUD Client
  Copyright (C) 2000 Ron Martin <rmartin@cps.cmich.edu>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <ctype.h>

#include "client.h"
#include "callbacks.h"
#include "interface.h"
#include "constants.h"
#include "helpers.h"

extern struct conn_info * info_database;
extern GdkColor colors[8];
extern GdkColor color_normal;

/*
	destroy() - function to destroy the main window and quit the application.
*/
gint destroy(GtkWidget * widget,GdkEvent  * event, gpointer data)
{
	gtk_object_destroy(GTK_OBJECT(widget));
	gtk_main_quit();
	return(FALSE);
}

/*
	delete_event() - function to delete a widget.
*/
gint delete_event(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	gtk_object_destroy(GTK_OBJECT(widget));
	return TRUE;
}

/*
	close_conn() - handels a click event on the close button of an open connection.
	The function needs to check to see if it is opened, and if it is, it deletes the open
	connection window and sets the connection's open state to FALSE.
*/
gint close_conn(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;
	struct log_info * temp_log;
	
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	
	if(temp_conn->state & STATE_CONNECTED)
	{
		error_handle(15);
		return TRUE;
	}
	
	temp_log = temp_conn->logs;
	while(temp_log)
	{
		delete_log(temp_log->log_widget, NULL, temp_log->log_widget);
		temp_log = temp_log->next;
	}
	delete_event(widget, NULL, NULL);
	temp_conn->state &= ~STATE_OPEN;
	return FALSE;
}

/*
	create_new_conn() - creates an open connection window and displays it.
	The function gets the list of connections from the `main' window and creates
	the New Connection dialog.  It adds the list's address to the data held by
	the New Connection window.
*/
gint create_new_conn(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * window_new_conn;
	GtkWidget * clist;
	
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist");
	
	window_new_conn = create_window_new_conn();
	gtk_object_set_data(GTK_OBJECT(window_new_conn), "clist", clist);
	gtk_widget_show(window_new_conn);

	return FALSE;
}

/*
	create_selection() - creates an open conenction window with the selected information.
	This function gets the user-selected connection information from the
	`main' window.  It checks to see if the connection is already opened and
	fails with an error message if it is.  If not, it continues to create the
	window.  It then creates a socket with the specified connection information.
	It also adds the STATE_CONNECTED flag to the connection's flags.
*/
gint create_selection(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * window = NULL, * help = NULL;
	int socket_descriptor;
	struct conn_info * temp_conn;
	char * title = NULL;
	struct log_info * log;
	void update_conn_text(gpointer output_widget, int socket_descriptor, GdkInputCondition condition);
	
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "selected_row");
	help = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "help_window");
	
	if(!temp_conn)
		return FALSE;
	if(temp_conn->state & STATE_OPEN)
	{
		error_handle(14);
		return FALSE;
	}
	
	window = create_window_conn(help);

	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "selected_row");
	title = g_strdup(temp_conn->mud_name);
	
	gtk_window_set_title(GTK_WINDOW(window), title);

	socket_descriptor = new_socket(temp_conn->host_name, temp_conn->port);
	if(socket_descriptor)
	{
		temp_conn->sock = socket_descriptor;
		temp_conn->descriptor = gdk_input_add(socket_descriptor, GDK_INPUT_READ, update_conn_text, window);
		gtk_object_set_data(GTK_OBJECT(window), "conn_info", temp_conn);
		temp_conn->state = (temp_conn->state | STATE_CONNECTED);
	}
	else
		return FALSE;
		
	temp_conn->state = temp_conn->state | STATE_OPEN;
	temp_conn->value = -1.0;

	gtk_object_set_data(GTK_OBJECT(window), "parent_window", widget);
	
	gtk_widget_show(window);
	
	log = temp_conn->logs;
	while(log)
	{
		if(log->log_to_window)
			log->log_widget = create_log_window(temp_conn->mud_name, log->name, log);
		log = log->next;
	}
	
	return FALSE;
}

/*
	edit_selection() - create the `edit' window to edit mud info, triggers, and aliases.
	This function gets the selected connection information and creates the edit window.
	It also checks to see if it is already being edited, and sets the STATE_EDITING flag
	when finished.
*/
gint edit_selection(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * window;
	struct conn_info * temp_conn;
	
	void init_edit_conn(GtkWidget * window, struct conn_info * temp_conn);
	
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "selected_row");
	
	if(temp_conn->state & STATE_EDITING)
	{
		error_handle(13);
		return FALSE;
	}
	
	window = create_window_edit_conn_chose();

	gtk_object_set_data(GTK_OBJECT(window), "parent_window", widget);
		
	init_edit_conn(window, temp_conn);
	
	gtk_widget_show(window);
	temp_conn->state = temp_conn->state | STATE_EDITING;
		
	return FALSE;
}

/*
	select_row() - callback for selecting a row in the `main' window.
	This function gets passed the row number selected, and iterates through
	the conenction linked list row times.  It then sets the selected_row data
	associated with the window to point to the connection.
*/
gint select_row(GtkWidget * widget, int row, int column, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_info;
	GtkWidget * clist;
	
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist");
	temp_info = (struct conn_info *)gtk_clist_get_row_data(GTK_CLIST(clist), row);
	
	gtk_object_set_data(GTK_OBJECT(widget), "selected_row", temp_info);
	return FALSE;
}

/*
	send_text() - callback for when the user gives a open connection a command.
	This function gets the connection information and calls 
	process_user_input() to process any aliases.
*/
gint send_text(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * entry, * window, * text;
	int socket_descriptor;
	char * input;
	struct conn_info * temp_conn;
	struct commands * temp_command;
	
	window = (GtkWidget *)widget;
	temp_conn = gtk_object_get_data(GTK_OBJECT(window), "conn_info");

	entry = gtk_object_get_data(GTK_OBJECT(window), "entry");
	socket_descriptor = temp_conn->sock;
	text = gtk_object_get_data(GTK_OBJECT(window), "text");
	
	input = gtk_entry_get_text(GTK_ENTRY(entry));

	/* This check was added 6/20/00 - RMC
	   It is meant to prevent any "hidden" text from
	   showing up in the command history.
	*/
	if(GTK_ENTRY(entry)->editable.visible)
	{
		temp_command = temp_conn->history;
		if(!temp_command)
		{
			/* No history list.  Make one. */
			temp_conn->history = g_malloc0(sizeof(commands));
			temp_command = temp_conn->history;
		}
		else
		{
			/* Find the end of the list append add this command */
			while(temp_command->next)
				temp_command = temp_command->next;
			temp_command->next = g_malloc0(sizeof(commands));
			temp_command->next->prev = temp_command;
			temp_command = temp_command->next;
		}
	
		temp_conn->history = temp_command;
	
		temp_command->command = g_strdup(input);

	}

	/* Do something with the text. */
	process_user_input(input, temp_conn->aliases, socket_descriptor, widget);
	
	/* So the user doesn't need to delete what he or she just typed */
	gtk_editable_select_region(GTK_EDITABLE(entry), 0, -1);

	return FALSE;
}

/*
	select_row_alias() - callback for when an alias is selected in the edit window
	This function sets the current alias data to the alias selected by the user.  It
	calls the helper function select_row_helper() to take care of things that it
	shares with select_row_trigger()
*/
gint select_row_alias(GtkWidget * widget, int row, int column, GdkEvent * event, gpointer data)
{
	struct parse_type * temp_parse;
	GtkWidget * clist;
	
	void select_row_helper(GtkWidget * widget, struct parse_type * temp_parse, int temp);

	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_alias");

	temp_parse = gtk_clist_get_row_data(GTK_CLIST(clist), row);

	gtk_object_set_data(GTK_OBJECT(widget), "curr_alias", temp_parse);
	select_row_helper(widget, temp_parse, 1);
	return FALSE;
}

/*
	select_row_trigger() - callback for selection of trigger in edit window.
	This function is basically the same as select_row_alias(), but focuses on
	triggers.
*/
gint select_row_trigger(GtkWidget * widget, int row, int column, GdkEvent * event, gpointer data)
{
	struct parse_type * temp_parse;
	GtkWidget * clist;
	
	void select_row_helper(GtkWidget * widget, struct parse_type * temp_parse, int temp);
	
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_trigger");
	
	temp_parse = gtk_clist_get_row_data(GTK_CLIST(clist), row);
			
	gtk_object_set_data(GTK_OBJECT(widget), "curr_trigger", temp_parse);
	select_row_helper(widget, temp_parse, 2);
	return FALSE;
}

/*
	update_alias_clist_from_entry() - a dummy callback that calls
	update_clist_from_entry(), passing 1 as the second argument denoting the
	alias list should be updated.
*/
gint update_alias_clist_from_entry(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	void update_clist_from_entry(GtkWidget * widget, int type);
	
	update_clist_from_entry(widget, 1);
	return FALSE;
}

/*
	update_trigger_clist_from_entry() - a dummy callback that calls
	update_clist_from_entry(), passing 2 as the second argument denoting the
	trigger list should be updated.
*/
gint update_trigger_clist_from_entry(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	void update_clist_from_entry(GtkWidget * widget, int type);
	
	update_clist_from_entry(widget, 2);
	return FALSE;
}

/*
	edit_alias_text_change() - callbakc for when the alias text box in
	the edit window is changed.
	This function basically updates the alias information for the
	edited connection.  It calls edit_text_change() to handle those things
	common between the alias and triggers.
*/
gint edit_alias_text_change(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * text;
	struct parse_type  * temp_parse;
	
	void edit_text_change(GtkWidget * widget, struct parse_type * input_parse, int type);
	
	temp_parse = (struct parse_type *)gtk_object_get_data(GTK_OBJECT(widget), "curr_alias");
	if(!temp_parse)
		return FALSE;
		
	text = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "text_alias");
	edit_text_change(text, temp_parse, 1);
	
	return FALSE;
}

/*
	edit_trigger_text_change() - same as the above function, but with
	emphasis on triggers.
*/
gint edit_trigger_text_change(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * text;
	struct parse_type  * temp_parse;
	
	void edit_text_change(GtkWidget * text, struct parse_type * input_parse, int type);
	
	temp_parse = (struct parse_type *)gtk_object_get_data(GTK_OBJECT(widget), "curr_trigger");
	if(!temp_parse)
		return FALSE;

	text = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "text_trigger");
	edit_text_change(text, temp_parse, 1);
	
	return FALSE;
}

/*
	cancel_edit() - abandons an edit session in case the user is not
	satisfied with the changes.
	This function basically frees the temporary connection info used
	in editing a connection.
*/
gint cancel_edit(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_info;
	void conn_free(struct conn_info * temp_conn);
	
	temp_info = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	gtk_object_remove_data(GTK_OBJECT(widget), "conn_info");
	conn_free(temp_info);
	temp_info = NULL;
	temp_info = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "real_conn_info");
	temp_info->state &= ~STATE_EDITING;
	
	delete_event(widget, NULL, NULL);

	return FALSE;
}

/*
	save_edit() - callback for when the save button on the edit window
	is clicked.
	This function dumps the "real" connection and makes the database point to
	the "fake" connection.  The ol' swictharoo.
*/
gint save_edit(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * real_info, * fake_info;
	GtkWidget * temp_widget;
	
	void conn_free(struct conn_info * temp_conn);
	void write_info_database();
	struct conn_info * conn_dup(struct conn_info * input_conn);
	void conn_copy(struct conn_info * dest, struct conn_info * source);
	void init_clist(GtkWidget * widget);
	
	real_info = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "real_conn_info");
	fake_info = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");

	conn_copy(real_info, fake_info);
	
	conn_free(fake_info); 

	temp_widget = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "parent_window");
	temp_widget = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(temp_widget), "clist");
	init_clist(temp_widget);
	
	delete_event(widget, NULL, NULL);

	real_info->state &= ~STATE_EDITING;	
	write_info_database();

	return FALSE;
}

/*
	edit_delete_alias() - callback for when the delete button is pushed on the edit window.
	This function removes and frees the alias selected and updates the alias linked list
	so that the data pointing to the deleted section will point to the next set of data.  It
	also calls edit_delete_helper() to handle common procedures for aliases and triggers.
*/
gint edit_delete_alias(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;
	struct parse_type * temp_parse, * parent;
	GtkWidget * temp_widget;
	
	void parse_free(struct parse_type * input_parse);
	void edit_delete_helper(struct conn_info * parent_conn, struct parse_type * parent_parse, struct parse_type * input_parse, int type);

	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	temp_parse = (struct parse_type *)gtk_object_get_data(GTK_OBJECT(widget), "curr_alias");
	temp_widget = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_alias");
	if(!temp_parse)
		return FALSE;
		
	parent = temp_conn->aliases;

	edit_delete_helper(temp_conn, parent, temp_parse, 1);	
	parse_free(temp_parse);
	gtk_object_remove_data(GTK_OBJECT(widget), "curr_alias");
	init_parse_clist(temp_conn->aliases, temp_widget);
	if(temp_conn->aliases)
		gtk_clist_select_row(GTK_CLIST(temp_widget), 0, 0);
		
	return FALSE;
}

/*
	edit_delete_trigger() - callback for when the delete button is pushed on the edit window
	This function operates basically the same as edit_delete_alias() but with emphasis on
	triggers.
*/
gint edit_delete_trigger(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;
	struct parse_type * temp_parse, * parent;
	GtkWidget * temp_widget;
	
	void parse_free(struct parse_type * input_parse);
	void edit_delete_helper(struct conn_info * parent_conn, struct parse_type * parent_parse, struct parse_type * input_parse, int type);
	
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	temp_parse = (struct parse_type *)gtk_object_get_data(GTK_OBJECT(widget), "curr_trigger");
	temp_widget = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_trigger");
	if(!temp_parse)
		return FALSE;

	parent = temp_conn->triggers;
		
	edit_delete_helper(temp_conn, parent, temp_parse, 2);
	parse_free(temp_parse);
	gtk_object_remove_data(GTK_OBJECT(widget), "curr_trigger");
	init_parse_clist(temp_conn->triggers, temp_widget);
	if(temp_conn->triggers)	
		gtk_clist_select_row(GTK_CLIST(temp_widget), 0, 0);
	
	return FALSE;
}

/*
	edit_add_alias() - callback for when the add button is pushed in the edit window
	This function uses edit_add_helper() to handle the event.  It simply passes
	the alias list and a '1' to denote that an alias is being passed.
*/
gint edit_add_alias(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;
	GtkWidget * clist;
	
	void edit_add_helper(struct conn_info * data, GtkWidget * clist, int type);
	
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_alias");
	edit_add_helper(temp_conn, clist, 1);
	
	return FALSE;
}

/*
	edit_add_trigger() - callback for when the add button is pushed in the edit window
	This function is the same as edit_add_alias() but sends the trigger list and a '2'
	to denote a trigger is being sent.
*/
gint edit_add_trigger(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;
	GtkWidget * clist;
	
	void edit_add_helper(struct conn_info * data, GtkWidget * clist, int type);
	
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist_trigger");
	edit_add_helper(temp_conn, clist, 2);
	
	return FALSE;
}

/*
	edit_general_entry_host_name() - callback for when the host name is changed
	in the edit window.
	This function updates the host name of the edited selection.
*/
gint edit_general_entry_host_name(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * entry;
	struct conn_info * temp_conn;
	
	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_edit_general_host");
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	g_free(temp_conn->host_name);
	temp_conn->host_name = NULL;
	temp_conn->host_name = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
	
	return FALSE;
}

/*
	edit_general_entry_mud_name() - callback for when the mud name is changed
	This funciton updates the mud (connection) name of the edited selection.
*/
gint edit_general_entry_mud_name(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * entry;
	struct conn_info * temp_conn;
	
	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_edit_general_mud");
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	g_free(temp_conn->mud_name);
	temp_conn->mud_name = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
	
	return FALSE;	
}

/*
	edit_general_entry_port() - callback for when the port changes in
	the edit window.
	This function changes the port number of the edited selection.
*/
gint edit_general_entry_port(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * entry;
	struct conn_info * temp_conn;
	
	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "entry_edit_general_port");
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	temp_conn->port = atoi(gtk_entry_get_text(GTK_ENTRY(entry))); /* port is an int */
	
	return FALSE;
}

/*
	delete_selection() - This function deletes a selection from the info_database.
	This function finds the selection to delete, sets the data before it in the
	linked list to point to the data after the data being deleted, and then frees
	the selection.  It then updates the list in the main window, selects the first
	element, and writes the connection database out to file.
*/
gint delete_selection(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * clist;
	struct conn_info * delete_conn, * temp_conn;
	
	void conn_free(struct conn_info * conn);
	void init_clist(GtkWidget * widget);
	void write_info_database();
	
	delete_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "selected_row");
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "clist");
	
	temp_conn = info_database;
	
	if(!temp_conn)
	{
		error_handle(10);
		return FALSE;
	}
	
	if(temp_conn == delete_conn) /* if the selection is the first one */
		info_database = delete_conn->next; /* don't bother looking for it. */
	else
	{
		while(temp_conn->next) /*find the selection*/
		{
			if(temp_conn->next == delete_conn)
				break;
			temp_conn = temp_conn->next;
		}
		
		if(!temp_conn->next) /* Something is broken. */
		{
			error_handle(11);
			return FALSE;
		}
		
		temp_conn->next = delete_conn->next;
	}
	
	conn_free(delete_conn);
	init_clist(clist);
	write_info_database();
	
	return FALSE;
}

/*
	reconnect() - callback for when the reconnect button is pushed.
	This function first checks to see if the connection is already
	connected.  If so, it will send out an error and exit.  If not,
	it will create a new socket connection to the connection and
	set the connection state to STATE_CONNECTED.
*/
gint reconnect(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_info;
	int socket_descriptor;
	
	int new_socket(char * host_name, int port);	
	
	temp_info = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	
	if(temp_info->state & STATE_CONNECTED)
	{
		error_handle(12);
		return FALSE;
	}
	
	socket_descriptor = new_socket(temp_info->host_name, temp_info->port);
	if(socket_descriptor)
	{
		temp_info->sock = socket_descriptor;
		temp_info->descriptor = gdk_input_add(socket_descriptor, GDK_INPUT_READ, update_conn_text, widget);
		temp_info->state = (temp_info->state | STATE_CONNECTED);
	}

	return FALSE;
}

/*
	disconnect_ask() - callback for when the disconnect button is pressed.
	This function creates a dialog that asks the user if it is okay to
	disconnect if and only if the connection is open.
*/
gint disconnect_ask(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_info;
	GtkWidget * dialog;
	
	temp_info = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	if(temp_info->state & STATE_CONNECTED)
	{
		dialog = create_disconnect_dialog();
		gtk_object_set_data(GTK_OBJECT(dialog), "conn_info", temp_info);
		gtk_widget_show_all(dialog);
	}
	else
		temp_info->state &= ~STATE_CONNECTED;
	
	return FALSE;
}

/*
	disconnect() - callback for when the users agrees to disconnect.
	This hangs up the connection, sets the STATE_CONNECTED flag to
	false, and deletes the 'disconnect_ask' window.
*/
gint disconnect(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_info;
	gint delete_event(GtkWidget * widget, GdkEvent * event, gpointer data);
	
	temp_info = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	
	hang_up(temp_info->descriptor, temp_info->sock);
	temp_info->state &= ~STATE_CONNECTED;
	delete_event(widget, NULL, NULL);
	
	return TRUE;
}

/*
	edit_selection_from_window() - callback for when the edit button
	is clicked from the conenction window as opposed to the main window.
	This function first makes sure that the connection is not already
	being edited so that the application doesn't mess up the connection
	database.  It also calls init_edit_conn() to initialize the edit winow
	with the proper connection information.  It then sets the STATE_EDITING
	flag.
*/
gint edit_selection_from_window(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * window, * temp;
	struct conn_info * temp_conn;
	
	void init_edit_conn(GtkWidget * window, struct conn_info * temp_conn);
	
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	
	if(temp_conn->state & STATE_EDITING)
	{
		error_handle(13);
		return FALSE;
	}
	
	window = create_window_edit_conn_chose();
	
	temp = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "parent_window");
	gtk_object_set_data(GTK_OBJECT(window), "parent_window", temp);

	init_edit_conn(window, temp_conn);
	
	gtk_widget_show(window);
		
	return FALSE;
}	

/*
	hide_help() - Hide the help window
	This hides a global help window.  This was done so that
	the user cannot create fifty help windows and also to make
	it faster to access the help, somthing I feel is important.
*/
gint hide_help(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	gtk_widget_hide(widget);
	return TRUE;
}

/*
	show_help() - shows the help window
	This function shows the help window.  See the above function's comments
	to find out why this was done.
*/
gint show_help(GtkWidget * widget, GdkEvent * evetn, gpointer data)
{
	gtk_widget_show(widget);
	return FALSE;
}

/*
	main_check_close() - check onnections before closing.
	This will go through the info database to find any connections still open.
	If a connection is open, then an error will appear and the application will
	NOT close.  Otherwise, it's business as usual.  This was done to eliminate
	the strange behavior of the application to leave orphan connection windows
	lying about after quitting.
*/
gint main_check_close(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_info = info_database;
	
	while(temp_info)
	{
		if(temp_info->state & STATE_CONNECTED)/* are we connected? */
		{
			error_handle(16);
			return TRUE;
		}
		temp_info = temp_info->next;
	}
	
	gtk_main_quit(); /* That's all, folks! */
	return TRUE;
}

/*
	delete_log() - deletes a log window
	This function nicely deletes a log window.
*/
gint delete_log(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct log_info * log;
	
	if(!widget)
		return FALSE;
		
	log = gtk_object_get_data(GTK_OBJECT(widget), "log");
	delete_event(widget, NULL ,NULL);
	log->log_widget = NULL;
	
	return TRUE;
}

/*
	update_log_clist_from_entry() - updates the clist text from an entry
	This is just eye candy that updates a log's clist text when the text in
	an entry is changed.
*/
gint update_log_clist_from_entry(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	int row;
	struct log_info * log;
	GtkWidget * clist, * entry;
	
	log = (struct log_info *)gtk_object_get_data(GTK_OBJECT(widget), "curr_log");
	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_name_entry");
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_clist");
	
	if(log)
	{
		row = gtk_clist_find_row_from_data(GTK_CLIST(clist), log);
		gtk_clist_set_text(GTK_CLIST(clist), row, 0, gtk_entry_get_text(GTK_ENTRY(entry)));
		free(log->name); /* garbage collection is your friend. */
		log->name = strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
	}
	
	return FALSE;
}

/*
	update_log_file() - updates the text for the log file
	This will update the text in a log structure for the file name.  It will
	not allow a space to be entered, for obvious reasons.
*/
gint update_log_file(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct log_info * log;
	GtkWidget * entry;
	int i = 0;
	
	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_file_entry");
	log = (struct log_info *)gtk_object_get_data(GTK_OBJECT(widget), "curr_log");
	
	if(log)
	{
		i = 0;
		g_free(log->log_file);
		log->log_file = strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
		while(log->log_file[i] != '\0' && log->log_file[i] != ' ')
			i++;
			
		if(log->log_file[i] == ' ')
		{
			log->log_file[i] = '\0';
			gtk_entry_set_text(GTK_ENTRY(entry), log->log_file);
		}
		else if(log->log_file[0] == '\0')
			log->log_file = NULL;
	}
	
	return FALSE;
}

/*
	log_window_toggled() - changes the log window variable when a toggle button changes.
	This will set or unset window logging for a log object when a toggle button
	is clicked.
*/
gint log_window_toggled(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct log_info * log;
	GtkWidget * toggle_button;
	
	toggle_button = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "toggle_button");
	log = (struct log_info *)gtk_object_get_data(GTK_OBJECT(widget), "curr_log");
	
	if(log)
		log->log_to_window = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle_button));
	
	return FALSE;
}

/*
	select_row_log() - Callback for user selection of a log
	This will adjust any log object information when selecting new logs.
*/
gint select_row_log(GtkWidget * widget, int row, int column, GdkEvent * event, gpointer data)
{
	struct log_info * log;
	GtkWidget * clist, * name_entry, * file_name, * toggle_button;
	
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_clist");
	name_entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_name_entry");
	file_name = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_file_entry");
	toggle_button = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "toggle_button");
	
	log = (struct log_info *)gtk_clist_get_row_data(GTK_CLIST(clist), row);
	
	if(log)
	{		
		gtk_object_set_data(GTK_OBJECT(widget), "curr_log", log);
		if(log->name)
			gtk_entry_set_text(GTK_ENTRY(name_entry), log->name);
		else
			gtk_entry_set_text(GTK_ENTRY(name_entry), "\0");
		if(log->log_file)
			gtk_entry_set_text(GTK_ENTRY(file_name), log->log_file);
		else
			gtk_entry_set_text(GTK_ENTRY(file_name), "\0");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(toggle_button), log->log_to_window);
	}
	return FALSE;
}

/*
	edit_add_log() - Add a log to the list
	This adds a log to the log list, giving it default values.
*/
gint edit_add_log(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * clist;
	struct conn_info * temp_conn;
	struct log_info * log;
	
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_clist");
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	
	if(temp_conn->logs)
	{
		log = temp_conn->logs;
		while(log->next)
			log = log->next;
		log->next = g_malloc0(sizeof(log_info));
		log = log->next;
	}
	else
	{
		temp_conn->logs = g_malloc0(sizeof(log_info));
		log = temp_conn->logs;
	}
	
	log->name = g_strdup("NewLog");
  	init_log_clist(temp_conn->logs, GTK_CLIST(clist));
	return FALSE;
}

/*
	edit_delete_log() - Deletes a log from the list
	This makes sure that all pointers to the log are handled appropriately.
	The log before the opne being deleted will point to the one after the
	log being deleted.  If the log being deleted is the first log, then
	the parent log of the list is updated.
*/
gint edit_delete_log(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * clist;
	struct conn_info * temp_conn;
	struct log_info * delete_log, * log;
	
	clist = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "log_clist");
	temp_conn = (struct conn_info *)gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	log = temp_conn->logs;
	delete_log = (struct log_info *)gtk_object_get_data(GTK_OBJECT(widget), "curr_log");
	
	if(!log)
		return FALSE;
		
	if(delete_log == log)
	{
		temp_conn->logs = delete_log->next;
		free_log(delete_log);
	}
	else
		while(log->next)
		{
			if(log->next == delete_log)
			{
				log->next = delete_log->next;
				free_log(delete_log);
				break;
			}
			log = log->next;
		}

	init_log_clist(temp_conn->logs, GTK_CLIST(clist));
	return FALSE;
}

/*
	scroll_command_history() - scroll forward or backward through the command history
	This will find the arrow key pushed.  If it is up arrow or down arrow, then the
	history will be scrolled through and the appropriate text will be put into the
	entry area.  The key_press signals will be stopped so that other widgets will not
	be "selected" by the event.
*/
gint scroll_command_history(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;
	GtkWidget * entry;
	int handled = FALSE;
	entry = gtk_object_get_data(GTK_OBJECT(widget), "entry");
	temp_conn = gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	
	switch(event->key.keyval)
	{
		case 0xFF52:
			handled = TRUE;
			if(temp_conn->history->prev)
			{
				temp_conn->history = temp_conn->history->prev;
				gtk_entry_set_text(GTK_ENTRY(entry), temp_conn->history->command);
				gtk_entry_select_region(GTK_ENTRY(entry), 0, -1);
			}
			break;
		case 0xFF54:
			handled = TRUE;
			if(temp_conn->history->next)
			{
				temp_conn->history = temp_conn->history->next;
				gtk_entry_set_text(GTK_ENTRY(entry), temp_conn->history->command);
				gtk_entry_select_region(GTK_ENTRY(entry), 0, -1);
			}
			break;
	}
	
	if (handled)
		gtk_signal_emit_stop_by_name (GTK_OBJECT(entry), "key_press_event");

	return handled;
}

/*
	update_font() - Updates the font from a font selection dialog.
	This will get the font name from the dialog box and then load the font.
*/
gint update_font(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;

	temp_conn = gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	g_free(temp_conn->font);
	g_free(temp_conn->font_name);
	temp_conn->font_name = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(widget));
	temp_conn->font = gdk_font_load(temp_conn->font_name);
	gtk_object_destroy(GTK_OBJECT(widget));
	
	return TRUE;
}

/*
	choose_font() - Callback for user wanting to change the font
	This function creates the font dialog and presents it to the user.
*/	
gint choose_font(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	struct conn_info * temp_conn;
	GtkWidget * font_select;
	
	temp_conn = gtk_object_get_data(GTK_OBJECT(widget), "conn_info");
	font_select = create_fontselectiondialog1(temp_conn);
	
	return FALSE;
}

/*
	list_row_select() - Callback for user selecting a mud from the list
	This will adjust any data, including the text in the "new connection" entry areas
	whenever the user selects a mud from the list.
*/
gint list_row_select(GtkWidget * widget, int row, int col, GdkEvent * event, gpointer data)
{
	GtkWidget * parent;
	GtkWidget * entry;
	char * text;
	
	parent = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(widget), "parent");
	
	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(parent), "entry_name_mud");
	gtk_clist_get_text(GTK_CLIST(widget), row, 0, &text);
	gtk_entry_set_text(GTK_ENTRY(entry), text);
	
	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(parent), "entry_name_host");
	gtk_clist_get_text(GTK_CLIST(widget), row, 1, &text);
	gtk_entry_set_text(GTK_ENTRY(entry), text);

	entry = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(parent), "entry_port");
	gtk_clist_get_text(GTK_CLIST(widget), row, 2, &text);
	gtk_entry_set_text(GTK_ENTRY(entry), text);

	return FALSE;
}

/*
	show_mud_list() - shows the mud list
*/
gint show_mud_list(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * window;

	window = create_mud_list(widget);
	init_mud_list(window);
	
	return FALSE;
}

/*
	update_mud_list() - Updates the MUD list from www.mudconnector.com
	This creates a new socket to www.mudconnector.com, port 80, and sends an
	HTML command to retrieve first the resource header (for the size) and
	then a request for the mud list.  The list widget is then updated to
	reflect the new results.
	
	NOTE: 	THE LIST ON WWW.MUDCONNECTOR.COM IS USED BY PERMISSION OF THE MUD
	CONNECTOR.  ANY OTHER USE OF THE LIST MAY BE IN VIOLATION OF COPYRIGHT.
	PLEASE CONTACT THE ADMINISTRATOR OF THE MUD CONNECTOR AT
		admin@mudconnector.com
	FOR INFORMATION ON THE TERMS OF USAGE OF THE LIST.
*/
int update_mud_list(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	GtkWidget * window;
	int * descriptor, * tag, * size;
	char * header, * temp, * temp2;
	
	tag = g_malloc(sizeof(int));
	descriptor = g_malloc(sizeof(int));
	size = g_malloc(sizeof(int));
	header = g_malloc0(2049);
	
	temp = header;
	
	*descriptor = new_socket("www.mudconnector.com", 80);

	if(!(*descriptor))
	{
		error_handle(21);
		return FALSE;
	}
	
	send(*descriptor, "HEAD /clients/mudlist2.txt HTTP/1.0\n\n", 37, 0);
	/* ^^ send an HTTP/1.0 request for the content header so we can get the resource size. */
	recv(*descriptor, header, 2048, 0);
	close(*descriptor);
	
	while(1) /* find the "Content-Length" string */
	{
		temp2 = strsep(&header, "C");
		if(!temp2)
		{
			error_handle(23);
			return FALSE;
		}
		if(!(strncmp(temp2, "ontent-Length:", 14)))
			break;
	}
	
	while(!(isdigit(*temp2)))
		temp2++;
		
	*size = atoi(temp2);
	
	g_free(temp);
	
	*descriptor = new_socket("www.mudconnector.com", 80);

	if(!(*descriptor))
	{
		error_handle(21);
		return FALSE;
	}
	
	send(*descriptor, "GET /clients/mudlist2.txt\n", 26, 0);
	
	system("rm ~/.rmclist");
	system("touch ~/.rmclist");
	
	window = create_download_window();
	
	*tag = gdk_input_add(*descriptor, GDK_INPUT_READ, get_list_data, window);
	
	gtk_object_set_data(GTK_OBJECT(window), "tag", tag);
	gtk_object_set_data_full(GTK_OBJECT(window), "descriptor", descriptor, g_free);
	gtk_object_set_data_full(GTK_OBJECT(window), "size", size, g_free);
	gtk_object_set_data(GTK_OBJECT(window), "parent", widget);
	
	return FALSE;
}

/*
	cancel_download() - Cancel the MUD list download.
	This will cancel the download and close the socket associated
	with the network transfer.
*/
gint cancel_download(GtkWidget * widget, GdkEvent * event, gpointer data)
{
	int * tag, * descriptor;
	
	tag = gtk_object_get_data(GTK_OBJECT(widget), "tag");
	descriptor = gtk_object_get_data(GTK_OBJECT(widget), "descriptor");
	
	hang_up_list_conn(*tag, *descriptor);
	
	g_free(tag);
	
	delete_event(widget, event, data);
	
	return TRUE;
}
